/*
PRICE was written by Graham Ruby.
This file is part of PRICE.

PRICE is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

PRICE is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with PRICE.  If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef EXTENDCYCLE_CPP
#define EXTENDCYCLE_CPP

#include "ExtendCycle.h"
#include "ExtendJobCreator.h"
#include "Read.h"
#include "ScoredSeqSubseq.h"
#include "ScoredSeqNested.h"
#include "ScoredSeqMonoScore.h"
#include <typeinfo>
#include <sstream>
#include <omp.h>
#include <limits.h>
using namespace::std;

int ExtendCycle::_plusIndex = 0;
int ExtendCycle::_minusIndex = 1;
int ExtendCycle::_frontIndex = 1;
int ExtendCycle::_backIndex = 0;

int ExtendCycle::plusIndex(){ return _plusIndex; }
int ExtendCycle::minusIndex(){ return _minusIndex; }
int ExtendCycle::frontIndex(){ return _frontIndex; }
int ExtendCycle::backIndex(){ return _backIndex; }


void ExtendCycle::constructorHelper(set<ParameterizedReadFile*>* readFileSet,
				    set<ScoredSeq*>* inputContigs,
				    set<ScoredSeq*>* doNotExtend){

  _cycleHasRun = false;

  _numReadFiles = readFileSet->size();
  _readFileArray = new ParameterizedReadFile*[ _numReadFiles ];
  _hitReadsByFile = new vector<ScoredSeqWithCarriers*>*[ _numReadFiles ];
  for (int n = 0; n < _numReadFiles; ++n){ _hitReadsByFile[n] = new vector<ScoredSeqWithCarriers*>; }

  int rfCounter = 0;
  for (set<ParameterizedReadFile*>::iterator rfIt = readFileSet->begin(); rfIt != readFileSet->end(); ++rfIt){
    _readFileArray[rfCounter] = (*rfIt);
    rfCounter++;
  }

  // make shallow copies of the input contigs so that they can be deleted
  for (set<ScoredSeq*>::iterator icIt = inputContigs->begin(); icIt != inputContigs->end(); ++icIt){
    //ScoredSeqNormalized* inSeq = new ScoredSeqNormalized( (*icIt)->shallowCopy() );
    ScoredSeqNormalized* inSeq = new ScoredSeqNormalized( *icIt );
    _inputContigs.insert( inSeq );
    if ( doNotExtend->count( (*icIt) ) != 0 ){ _doNotExtendContigs.insert( inSeq ); }
  }

  _numThreads = omp_get_max_threads();
}

ExtendCycle::ExtendCycle(int cycleNum, set<ParameterizedReadFile*>* readFileSet, ReadPairFilter* rpfSingle, ReadPairFilter* rpfDouble, set<ScoredSeq*>* inputContigs, AssemblerListener* listener) :
  _cycleNum(cycleNum),
  _listener(listener) {
  _rpfSingle = rpfSingle->copy();
  _rpfDouble = rpfDouble->copy();
  set<ScoredSeq*> emptyDoNotExtend;
  constructorHelper(readFileSet, inputContigs, &emptyDoNotExtend);
}

ExtendCycle::ExtendCycle(int cycleNum, set<ParameterizedReadFile*>* readFileSet, ReadPairFilter* rpfSingle, ReadPairFilter* rpfDouble, set<ScoredSeq*>* inputContigs, set<ScoredSeq*>* doNotExtend, AssemblerListener* listener) :
  _cycleNum(cycleNum),
  _listener(listener) {
  _rpfSingle = rpfSingle->copy();
  _rpfDouble = rpfDouble->copy();
  constructorHelper(readFileSet, inputContigs, doNotExtend);
}

ExtendCycle::~ExtendCycle(){
  delete _rpfSingle;
  delete _rpfDouble;
  delete [] _readFileArray;
  for (int n = 0; n < _numReadFiles; ++n){ delete _hitReadsByFile[n]; }
  delete [] _hitReadsByFile;

  // get rid of the wrappers of the contigs (the map set outermost wrappers)
  if (_mapSets.size() != 0){
    throw AssemblyException::ImplementationError("EC:destructor, mapSets still exist; these should have already been deleted."); 
  }
}

void ExtendCycle::runCycle(ParamsAlignment * paMini,
			   ParamsAlignment * paMeta,
			   ParamsDeBruijn* pdb,
			   long maxContigLinkages,
			   EcoFilter* preMetaFilter,
			   EcoFilter* fullFilter){

  // set up the carriers that will hold the read-contig mapping linkages
  set<ScoredSeqWithCarriers*> dneWithCarriers;
  for (set<ScoredSeqNormalized*>::iterator it = _inputContigs.begin(); it != _inputContigs.end(); ++it){
    ScoredSeqNormalized* contig = (*it); // the contig will be normalized
    ScoredSeqWithCarriers* carrier = new ScoredSeqWithCarriers(contig, 2); // 0->frontSet, 1->backSet
    _mapSets.insert( carrier );
    if ( _doNotExtendContigs.count(contig) != 0){ dneWithCarriers.insert(carrier); }
  }

  // first, map to the contig edges of "alive" contigs to establish the
  // set of relevant reads
  MapReadSetupHelper* mrsh;
  ParameterizedReadFile** nullRfa = new ParameterizedReadFile*[1];
  nullRfa[0] = NULL;
  if (dneWithCarriers.size() == _inputContigs.size()){
    mrsh = new MrsMainHelper(nullRfa, 0, _rpfDouble, &dneWithCarriers);
  } else {
    mrsh = new MrsMainHelper(_readFileArray, _numReadFiles, _rpfDouble, &dneWithCarriers);
  }
  mapReads(mrsh);
  delete mrsh;

  // second, map the relevant reads to ALL of the contigs.  "relevant" includes
  // the reads that matched in the first step plus their paired ends.  first,
  // the matches from the first step must be erased; they will either be re-discovered
  // or replaced by better-scoring matches to other parts of the assembly
  for (int n = 0; n < _numReadFiles; ++n){ clearLinkages(_hitReadsByFile[n]); }
  mrsh = new MrsAddendumHelper(_readFileArray, _numReadFiles, _rpfSingle, _hitReadsByFile);
  mapReads(mrsh);
  delete mrsh;

  // now, perform the assemblies
  runMiniAssemblies(paMini, pdb, maxContigLinkages, preMetaFilter);
  runMetaAssembly(paMeta, pdb);
  // use the full filter to filter the potential output
  _metaInputContigs.clear();
  set<ScoredSeq*> failingContigs;
  fullFilter->filterSeqs(&_assembledContigs, &_metaInputContigs, &failingContigs, _cycleNum, EcoFilter::threaded);
  // i don't need to do anything if there were no sequences filtered out
  if (failingContigs.size() != 0){
    _assembledContigs.clear();
    for (set<ScoredSeq*>::iterator it = failingContigs.begin(); it != failingContigs.end(); ++it){ (*it)->deepDelete(); }
    runMetaAssembly(paMeta, pdb);
  }

  delete [] nullRfa;
  _cycleHasRun = true;
}



void ExtendCycle::clearLinkages(vector<ScoredSeqWithCarriers*>* linkedSeqs){
  for (vector<ScoredSeqWithCarriers*>::iterator it = linkedSeqs->begin(); it != linkedSeqs->end(); ++it){
    (*it)->clearCarriedSeqs(0);
    (*it)->clearCarriedSeqs(1);
  }
}





ExtendCycle::MapReadSetupHelper::~MapReadSetupHelper(){}


ExtendCycle::MrsMainHelper::MrsMainHelper(ParameterizedReadFile** readFileArray, int numReadFiles, ReadPairFilter* rpf, set<ScoredSeqWithCarriers*>* doNotExtendContigs) :
  _numReadFiles(numReadFiles),
  _rpf(rpf) {// = rpf->copy();
  // figure out the total file sizes for stdout update and set up for threaded file reading
  _numActiveFiles = 0;
  _totalFileSize = 0;
  _doNotExtendContigs.insert(doNotExtendContigs->begin(), doNotExtendContigs->end());

  _readFileArray = new ParameterizedReadFile*[ _numReadFiles + 1 ];
  for (int n = 0; n < _numReadFiles; ++n){ _readFileArray[n] = readFileArray[n]; }
  _fileIsActive = new bool[ _numReadFiles ]; // keeps track of how many files are still being read
  _fileToPmap = new ParamsMapping*[ _numReadFiles ];

  #pragma omp parallel for schedule(dynamic)
  for (int n = 0; n < _numReadFiles; ++n){
    long numReads = _readFileArray[n]->numReads();
    //#pragma omp critical (ECmapReads)
    {
      #pragma omp atomic
      _totalFileSize += numReads;
      // pre-open the files that have reads
      if (numReads == 0){
        #pragma omp critical (ECmapReads)
	{
	  _fileIsActive[n] = false;
	  _fileToPmap[n] = 0;
	}
      } else {
        #pragma omp critical (ECmapReads)
	{
	  _fileToPmap[n] = _readFileArray[n]->getMapParams();
	  _fileIsActive[n] = true;
	}
        #pragma omp atomic
	_numActiveFiles++;
	_readFileArray[n]->openNormWithCarriers(2,false);
      }
    }
  }
}
ExtendCycle::MrsMainHelper::~MrsMainHelper(){
  for (int n = 0; n < _numReadFiles; ++n){ 
    _readFileArray[n]->close();
    delete _fileToPmap[n];
  }
  delete [] _readFileArray;
  delete [] _fileIsActive;
  delete [] _fileToPmap;
}

int ExtendCycle::MrsMainHelper::numFilesTotal(){ return _numReadFiles; }
int ExtendCycle::MrsMainHelper::numFilesActive(){ return _numActiveFiles; }
bool ExtendCycle::MrsMainHelper::isFileActive(int fileNum){ return _fileIsActive[fileNum]; }
void ExtendCycle::MrsMainHelper::triggerFiles(){
  for (int n = 0; n < _numReadFiles; ++n){
    ReadFileCommunicator::bufferFill(_readFileArray[n], ReadFile::bufferThreaded);
  }
}


long ExtendCycle::MrsMainHelper::totalReadCount(){ return _totalFileSize; }
ExtendJobMapper* ExtendCycle::MrsMainHelper::getNewEjm(set<ScoredSeqWithCarriers*>* parallelMapSet, int fileNum){
  return new ExtendJobMapperNoLinks( parallelMapSet, _fileToPmap[fileNum]->fractId(), _readFileArray[fileNum]->ampliconSize() );
}

ScoredSeqWithCarriers*** ExtendCycle::MrsMainHelper::getNewReadArrays(int* fileNumArray, int numFiles){
  ScoredSeqWithCarriers*** localReadArray1 = new ScoredSeqWithCarriers**[ numFiles + 1 ];
  ScoredSeqWithCarriers*** localReadArray2 = new ScoredSeqWithCarriers**[ numFiles + 1 ];
  long* localInitialReadCount = new long[ numFiles + 1 ];

  // get all the reads, threaded by file
  #pragma omp parallel for schedule(dynamic)
  for (int n = 0; n < numFiles; ++n){
    int fileNum = fileNumArray[n];
    ParameterizedReadFile* readFile = _readFileArray[fileNum];
    long arrayTargetSize = _fileToPmap[fileNum]->stepSize() / 2;
    if (arrayTargetSize == 0){ arrayTargetSize = 1; }
    long readCounter = 0;
    ScoredSeqWithCarriers** readArray1 = new ScoredSeqWithCarriers*[ arrayTargetSize + 1 ];
    ScoredSeqWithCarriers** readArray2 = new ScoredSeqWithCarriers*[ arrayTargetSize + 1 ];
    while ( readCounter < arrayTargetSize and readFile->hasRead() ){
      ScoredSeqWithCarriers* newRead1 = dynamic_cast<ScoredSeqWithCarriers*>( readFile->getRead() );
      // only paired-end reads will be mapped, if the file has a mixture
      if ( newRead1->hasPairedEnd() ){
	ScoredSeqWithCarriers* newRead2 = dynamic_cast<ScoredSeqWithCarriers*>( newRead1->getPairedEnd() );
	readArray1[readCounter] = newRead1;
	readArray2[readCounter] = newRead2;
	newRead1->bottomBuffer();
	newRead2->bottomBuffer();
	readCounter += 1;
      }
    }
    localReadArray1[n] = readArray1;
    localReadArray2[n] = readArray2;
    localInitialReadCount[n] = readCounter;
  }

  // buffer the reads for the pair test
  for (int n = 0; n < numFiles; ++n){
    int fileNum = fileNumArray[n];
    ReadFileCommunicator::bufferFill(_readFileArray[fileNum], ReadFile::bufferThreaded);
  }

  // now filter the results based on the pair test, threaded by the actual runnings of the test
  ScoredSeqWithCarriers*** allSeqsToMapArrays = new ScoredSeqWithCarriers**[ numFiles + 1 ];
  for (int n = 0; n < numFiles; ++n){
    int fileNum = fileNumArray[n];
    long readCounter = 0;
    ScoredSeqWithCarriers** seqsToMapArray = new ScoredSeqWithCarriers*[ localInitialReadCount[n] * 2 + 1 ];
    #pragma omp parallel for schedule(static)
    for (long readN = 0; readN < localInitialReadCount[n]; ++readN){
      ScoredSeqWithCarriers* newRead1 = localReadArray1[n][readN];
      ScoredSeqWithCarriers* newRead2 = localReadArray2[n][readN];
      if ( _rpf->isPairOk(newRead1, newRead2) ){
        #pragma omp critical (ECmrmhGnra)
	{
	  seqsToMapArray[ readCounter ] = newRead1;
	  seqsToMapArray[ readCounter+1 ] = newRead2;
	  readCounter += 2;
	}
      } else {
	newRead1->deepDelete();
	newRead2->deepDelete();
      }
    }
    // end each array with a null
    seqsToMapArray[readCounter] = NULL;
    allSeqsToMapArrays[n] = seqsToMapArray;
    delete [] localReadArray1[n];
    delete [] localReadArray2[n];
  }
  delete [] localReadArray1;
  delete [] localReadArray2;
  delete [] localInitialReadCount;
  return allSeqsToMapArrays;
}

void ExtendCycle::MrsMainHelper::updateActiveFiles(){
  _numActiveFiles = 0;
  for (int n = 0; n < _numReadFiles; ++n){
    if (_fileIsActive[n]){
      if ( _readFileArray[n]->hasRead() ){ _numActiveFiles++; }
      else { _fileIsActive[n] = false; }
    }
  }
}
bool ExtendCycle::MrsMainHelper::includeContigCarrier(ScoredSeqWithCarriers* carrier){ return _doNotExtendContigs.count(carrier) == 0; }






ExtendCycle::MrsAddendumHelper::MrsAddendumHelper(ParameterizedReadFile** readFileArray, int numReadFiles, ReadPairFilter* rpf, vector<ScoredSeqWithCarriers*>** hitReadsByFile) :
  _rpf(rpf),
  _numReadFiles(numReadFiles) {
  _numActiveFiles = 0;
  _totalNumQueries = 0;

  _readFileArray = new ParameterizedReadFile*[ _numReadFiles + 1 ];
  for (int n = 0; n < _numReadFiles; ++n){ _readFileArray[n] = readFileArray[n]; }

  // make sets of the paired-ends of the mapped reads, THOSE will be mapped
  _pairQueries = new set<ScoredSeqWithCarriers*>*[ _numReadFiles ];
  for (int n = 0; n < _numReadFiles; ++n){
    _pairQueries[n] = new set<ScoredSeqWithCarriers*>;
    for (vector<ScoredSeqWithCarriers*>::iterator it = hitReadsByFile[n]->begin(); it != hitReadsByFile[n]->end(); ++it){
      _pairQueries[n]->insert( dynamic_cast<ScoredSeqWithCarriers*>( (*it)->getPairedEnd() ) );
      // I also want to re-map the read to the entire dataset so that its repetitive nature (ability
      // to map to many different contigs) can be fairly evaluated
      _pairQueries[n]->insert( *it );
    }
  }

  // set up the iterators that will be used to allocate the reads
  _readItByFileNum = new set<ScoredSeqWithCarriers*>::iterator[ _numReadFiles ];
  _fileIsActive = new bool[ _numReadFiles ];
  _fileToPmap = new ParamsMapping*[ _numReadFiles ];
  for (int n = 0; n < _numReadFiles; ++n){
    _readItByFileNum[n] = _pairQueries[n]->begin();
    _totalNumQueries += _pairQueries[n]->size();
    _fileToPmap[n] = readFileArray[n]->getMapParams();
    if (_readItByFileNum[n] == _pairQueries[n]->end()){
      _fileIsActive[n] = false;
    } else {
      _fileIsActive[n] = true;
      _numActiveFiles++;
    }
  }
}
ExtendCycle::MrsAddendumHelper::~MrsAddendumHelper(){
  for (int n = 0; n < _numReadFiles; ++n){ 
    delete _pairQueries[n];
    delete _fileToPmap[n];
  }
  delete [] _pairQueries;
  delete [] _readItByFileNum;
  delete [] _fileIsActive;
  delete [] _fileToPmap;
  delete [] _readFileArray;
}
int ExtendCycle::MrsAddendumHelper::numFilesTotal(){ return _numReadFiles; }
int ExtendCycle::MrsAddendumHelper::numFilesActive(){ return _numActiveFiles; }
long ExtendCycle::MrsAddendumHelper::totalReadCount(){ return _totalNumQueries; }
bool ExtendCycle::MrsAddendumHelper::isFileActive(int fileNum){ return _fileIsActive[fileNum]; }
ExtendJobMapper* ExtendCycle::MrsAddendumHelper::getNewEjm(set<ScoredSeqWithCarriers*>* parallelMapSet, int fileNum){
  return new ExtendJobMapperWithLinks( parallelMapSet, _fileToPmap[fileNum]->fractId() );
}
ScoredSeqWithCarriers*** ExtendCycle::MrsAddendumHelper::getNewReadArrays(int* fileNumArray, int numFiles){
  ScoredSeqWithCarriers*** localReadArrays = new ScoredSeqWithCarriers**[ numFiles + 1 ];
  long* localInitialReadCount = new long[ numFiles + 1 ];
  #pragma omp parallel for schedule(dynamic)
  for (int n = 0; n < numFiles; ++n){
    int fileNum = fileNumArray[n];
    ParamsMapping* pmap = _fileToPmap[ fileNum ];
    ScoredSeqWithCarriers** readArray = new ScoredSeqWithCarriers*[ pmap->stepSizeAddendum() + 1 ];
    long readCounter = 0;
    while ( readCounter < pmap->stepSizeAddendum() and _readItByFileNum[ fileNum ] != _pairQueries[ fileNum ]->end() ){
      ScoredSeqWithCarriers* newRead = *_readItByFileNum[ fileNum ];
      readArray[ readCounter ] = newRead;
      readCounter += 1;
      newRead->bottomBuffer();
      ++_readItByFileNum[ fileNum ];
    }
    localReadArrays[n] = readArray;
    localInitialReadCount[n] = readCounter;
  }

  // buffer the reads for the pair test
  for (int n = 0; n < numFiles; ++n){
    int fileNum = fileNumArray[n];
    ReadFileCommunicator::bufferFill(_readFileArray[fileNum], ReadFile::bufferThreaded);
  }

  // now filter the results based on the pair test, threaded by the actual runnings of the test
  ScoredSeqWithCarriers*** allReadArrays = new ScoredSeqWithCarriers**[ numFiles + 1 ];
  //#pragma omp parallel for schedule(static)
  for (int n = 0; n < numFiles; ++n){
    int fileNum = fileNumArray[n];

    bool* isReadOkArray = new bool[ localInitialReadCount[n]+1 ];
    #pragma omp parallel for schedule(static)
    for (long readN = 0; readN < localInitialReadCount[n]; ++readN){
      isReadOkArray[readN] = _rpf->isReadOk(localReadArrays[n][readN]);
    }

    // don't thread this because paired reads could get deleted twice
    long readCounter = 0;
    ScoredSeqWithCarriers** seqsToMapArray = new ScoredSeqWithCarriers*[ localInitialReadCount[n] * 2 + 1 ];
    for (long readN = 0; readN < localInitialReadCount[n]; ++readN){
      if ( isReadOkArray[readN] ){
	seqsToMapArray[ readCounter ] = localReadArrays[n][readN];
	readCounter += 1;
      } else { localReadArrays[n][readN]->deepDelete(); }
    }
    delete [] isReadOkArray;
    // end each array with a null
    seqsToMapArray[readCounter] = NULL;
    allReadArrays[n] = seqsToMapArray;
    delete [] localReadArrays[n];
  }
  delete [] localReadArrays;
  delete [] localInitialReadCount;

  return allReadArrays;
}
void ExtendCycle::MrsAddendumHelper::triggerFiles(){
  for (int n = 0; n < _numReadFiles; ++n){
    ReadFileCommunicator::bufferFill(_readFileArray[n], ReadFile::bufferThreaded);
  }
}


void ExtendCycle::MrsAddendumHelper::updateActiveFiles(){
  _numActiveFiles = 0;
  for (int fileN = 0; fileN < _numReadFiles; ++fileN){
    if (_fileIsActive[fileN]){
      if (_readItByFileNum[fileN] == _pairQueries[fileN]->end()){
	_fileIsActive[fileN] = false;
      } else { _numActiveFiles++; }
    }
  }
}
bool ExtendCycle::MrsAddendumHelper::includeContigCarrier(ScoredSeqWithCarriers* carrier){ return true; }





void ExtendCycle::mapReads(MapReadSetupHelper* mrsh){

  // no need to do anything if there are no active files
  if (mrsh->numFilesActive() == 0){
    _listener->beginMapping(0);
    _listener->endMapping();
  } else {

    // set up for mapping - one EJM for each file, for each thread
    _listener->verboseMessage("Creating mapper objects...");
    ExtendJobMapper** fileToEjm = new ExtendJobMapper*[ _numReadFiles ];
    #pragma omp parallel for schedule(dynamic)
    for (int fileN = 0; fileN < mrsh->numFilesTotal(); ++fileN){
      if ( mrsh->isFileActive(fileN) ){
	fileToEjm[ fileN ] = mrsh->getNewEjm(&_mapSets, fileN);
      } else { fileToEjm[ fileN ] = 0; }
    }
    _listener->beginMapping(mrsh->totalReadCount());

    while (mrsh->numFilesActive() > 0){
      // prepare a subset of reads for mapping - fill these in below
      ScoredSeqWithCarriers*** seqsToMapArray = new ScoredSeqWithCarriers**[ _numThreads ];
      long seqsToMapCount[_numThreads];
      long seqsWithHitsCount[_numThreads];

      // get the files to use this round, the max num of files allowed by the num threads
      set<int> thisRoundFileNums;
      int fn = 0;
      while (fn < mrsh->numFilesTotal() and thisRoundFileNums.size() < _numThreads){
	if ( mrsh->isFileActive(fn) ){ thisRoundFileNums.insert(fn); }
	++fn;
      }

      // allocate threads to each file
      int numRoundFiles = thisRoundFileNums.size();
      int numThreadsUsed = 0;
      int threadToFileNum[_numThreads];

      // set up an array of targets
      ExtendJobMapper* ejmArray[_numThreads ];

      _listener->verboseMessage("getting and buffering reads...");
      while (numThreadsUsed < _numThreads and thisRoundFileNums.size() > 0){
        // get the files ready for read extraction by creating local thread arrays
	int numLocalThreads = thisRoundFileNums.size();
	if (numLocalThreads + numThreadsUsed > _numThreads){ numLocalThreads = _numThreads - numThreadsUsed; }
	int localIndex = 0;
	for (set<int>::iterator it = thisRoundFileNums.begin(); it != thisRoundFileNums.end(); ++it){
	  if (localIndex < numLocalThreads){
	    threadToFileNum[localIndex + numThreadsUsed] = *it;
	    localIndex++;
	  }
	}

        // get the queries for each thread (threaded by file) - do not thread this, the mrsh will thread it
	int* localFileN = new int[ numLocalThreads + 1 ];
	int* localThreadN = new int[ numLocalThreads + 1 ];
	for (int localN = 0; localN < numLocalThreads; ++localN){
  	  // get all of the necessary local variables
	  int threadN = localN + numThreadsUsed;
	  int fileN = threadToFileNum[threadN];
	  ejmArray[threadN] = fileToEjm[fileN];
	  localFileN[localN] = fileN;
	  localThreadN[localN] = threadN;
	}
	ScoredSeqWithCarriers*** localSeqsToMap = mrsh->getNewReadArrays(localFileN, numLocalThreads);
	for (int n = 0; n < numLocalThreads; ++n){
	  seqsToMapArray[localThreadN[n]] = localSeqsToMap[n];
	  long readCounter = 0;
	  while (seqsToMapArray[localThreadN[n]][readCounter] != 0){ readCounter++; }
	  seqsToMapCount[ localThreadN[n] ] = readCounter;
	}
	delete [] localSeqsToMap;
	delete [] localFileN;
	delete [] localThreadN;
	numThreadsUsed += numLocalThreads;

        // get rid of the finished file nums
	mrsh->updateActiveFiles();
	set<int> fileNumsToRemove;
	for (set<int>::iterator it = thisRoundFileNums.begin(); it != thisRoundFileNums.end(); ++it){
	  if (! mrsh->isFileActive(*it)){ fileNumsToRemove.insert( *it ); }
	}
	for (set<int>::iterator it = fileNumsToRemove.begin(); it != fileNumsToRemove.end(); ++it){ thisRoundFileNums.erase( *it ); }
      }

      // trigger the buffering of sequences OUTSIDE the parallel block
      _listener->verboseMessage("triggering files...");
      mrsh->triggerFiles();

      // do the mapping
      long readsBeingMapped = 0;
      for (int n = 0; n < numThreadsUsed; ++n){ readsBeingMapped += seqsToMapCount[n]; }

      _listener->verboseMessage("mapping reads...");
      #pragma omp parallel for schedule(dynamic)
      for (int threadN = 0; threadN < numThreadsUsed; ++threadN){
	long returnedCount = ejmArray[ threadN ]->mapQueries( seqsToMapArray[ threadN ], seqsToMapCount[ threadN ] );
	seqsWithHitsCount[ threadN ] = returnedCount;
	int fileN = threadToFileNum[ threadN ];
        #pragma omp critical (updateReturnCountArray)
	{ for (long n = 0; n < returnedCount; ++n){ _hitReadsByFile[ fileN ]->push_back( seqsToMapArray[ threadN ][n] ); } }
      }
      _listener->updateMapping(readsBeingMapped);
      for (int n = 0; n < numThreadsUsed; ++n){ delete [] seqsToMapArray[n]; }

      // clean up the EJMapper objects for non-active files.  I will set them
      // to zero so that I can still call delete on them in later mapping rounds.
      _listener->verboseMessage("cleaning up...");
      mrsh->updateActiveFiles();
      for (int n = 0; n < mrsh->numFilesTotal(); ++n){
	if (! mrsh->isFileActive(n)){
	  delete fileToEjm[n];
	  fileToEjm[n] = 0;
	}
      }
      delete [] seqsToMapArray;
    }
    delete [] fileToEjm;

    _listener->endMapping();
  }
}








void ExtendCycle::addBufferedSeqs(set<ScoredSeq*>* inputSet){
  #pragma omp critical (ECManageBuffer)
  {
    for (set<ScoredSeq*>::iterator seqIt = inputSet->begin(); seqIt != inputSet->end(); ++seqIt){
      map<ScoredSeq*,long>::iterator foundIt = _seqToBufferUseCount.find( *seqIt );
      if (foundIt == _seqToBufferUseCount.end()){
	_seqToBufferUseCount.insert( pair<ScoredSeq*,long>(*seqIt,1) );
	(*seqIt)->bottomBuffer();
      } else { foundIt->second++; }
    }
  }
}
void ExtendCycle::removeBufferedSeqs(set<ScoredSeq*>* inputSet){
  #pragma omp critical (ECManageBuffer)
  {
    for (set<ScoredSeq*>::iterator seqIt = inputSet->begin(); seqIt != inputSet->end(); ++seqIt){
      map<ScoredSeq*,long>::iterator foundIt = _seqToBufferUseCount.find( *seqIt );
      if (foundIt == _seqToBufferUseCount.end()){
	throw AssemblyException::LogicError("EC::removeBufferedSeqs cannot remove a non-existant buffered seq");
      }
      if (foundIt->second == 0){
	throw AssemblyException::LogicError("EC::removeBufferedSeqs cannot remove a zero-count buffered seq");
      }
      foundIt->second--;
      if (foundIt->second == 0){
	_seqToBufferUseCount.erase( foundIt );
	(*seqIt)->deepUnbuffer();
      }
    }
  }
}



void ExtendCycle::runOneMiniAssembly(AssemblyJob* aj, set<ScoredSeq*>* metaInputByThread, EcoFilter* preMetaFilter,
				     AssemblyJob::AssemblyThreadedness threadedness){
  // the input params of this factory need to be further utilized
  int threadNum = omp_get_thread_num();

  ostringstream outMessage1;
  outMessage1 << "Thread " << threadNum << ": Job Start";
  _listener->verboseMessage(outMessage1.str());

  // generate the buffer-threaded enum data type
  ReadFile::BufferThreadedness isBufferThreaded;
  if (threadedness == AssemblyJob::THREADED){ isBufferThreaded = ReadFile::bufferThreaded; }
  else { isBufferThreaded = ReadFile::bufferNotThreaded; }

  // make shallow copies of everything so that they can be deleted without worry
  set<ScoredSeq*> rawInput;
  aj->allInputSeqs( &rawInput );
  addBufferedSeqs( &rawInput );
  for (int n = 0; n < _numReadFiles; ++n){ ReadFileCommunicator::bufferFill(_readFileArray[n], isBufferThreaded); }
  aj->makeShallow(false);
  _listener->assemblyJobBeginStats(aj);
  removeBufferedSeqs( &rawInput );
  deleteReads( &rawInput );

  aj->runJob(threadedness);
  set<ScoredSeq*> unfilteredOutput;
  aj->allAssembledSeqs( &unfilteredOutput );

  // get rid of the too-$hort sequences
  set<ScoredSeq*> filteredMetaInput;
  set<ScoredSeq*> unsuitableMetaInput;

  // get the input data that was combined into the output contigs and delete it
  set<ScoredSeq*> discardedInput;
  aj->discardedInputSeqs( &discardedInput );
  for (set<ScoredSeq*>::iterator disIt = discardedInput.begin(); disIt != discardedInput.end(); ++disIt){ (*disIt)->deepDelete(); }

  _listener->assemblyJobEndStats(aj);
  ostringstream outMessage2;
  outMessage2 << "Thread " << threadNum << ": Job End";
  _listener->verboseMessage(outMessage2.str());
  _listener->updateEdgeAssembly(1);

  // deal with output
  if (isBufferThreaded){
    preMetaFilter->filterSeqs(&unfilteredOutput, metaInputByThread, &unsuitableMetaInput, _cycleNum, EcoFilter::threaded);
  } else {
    preMetaFilter->filterSeqs(&unfilteredOutput, metaInputByThread, &unsuitableMetaInput, _cycleNum, EcoFilter::notThreaded);
  }
  for (set<ScoredSeq*>::iterator cit = unsuitableMetaInput.begin(); cit != unsuitableMetaInput.end(); ++cit){ (*cit)->deepDelete(); }
}

void ExtendCycle::deleteReads(set<ScoredSeq*>* inputSet){
  #pragma omp critical (ECDeleteReads)
  {
    for (set<ScoredSeq*>::iterator seqIt = inputSet->begin(); seqIt != inputSet->end(); ++seqIt){
      // now delete, if used
      long totalCount = dynamic_cast<ScoredSeqNormalized*>( *seqIt )->getCount();
      map<ScoredSeq*,long>::iterator foundIt = _seqToTotalUseCount.find( *seqIt );
      if (foundIt != _seqToTotalUseCount.end()){
	foundIt->second++;
	if (foundIt->second == totalCount){
	  (*seqIt)->deepDelete();
	  _seqToTotalUseCount.erase( foundIt );
	}
      }
    }
  }
}


void ExtendCycle::runMiniAssemblies(ParamsAlignment* pa, ParamsDeBruijn* pdb, long maxContigLinkages, EcoFilter* preMetaFilter){

  AssemblyJobFactory * overhangFactory = new AssemblyJobFactory(pa, pdb, AssemblyJob::SINGLESTRANDED, _listener); //single-stranded

  // CREATE THE JOBS (this function contains threaded sections)
  set<ScoredSeqWithCarriers*> garbageDump;
  set<AssemblyJob*>* jobSet = ExtendJobCreator::createJobs(&_mapSets, &_doNotExtendContigs,
							   overhangFactory, maxContigLinkages,
							   &garbageDump);
  delete overhangFactory;
  // get rid of the wrappers of the contigs (the map set outermost wrappers)
  for (set<ScoredSeqWithCarriers*>::iterator inIt = _mapSets.begin(); inIt != _mapSets.end(); ++inIt){
    delete (*inIt); // shallow delete since inner contigs were already deleted
  }
  _mapSets.clear();

  // assembly jobs have been constructed, so carrier objects can be deleted and
  // nested objects stopred for future deletion
  set<ScoredSeq*> nestedGarbageDump;
  for (set<ScoredSeqWithCarriers*>::iterator it = garbageDump.begin(); it != garbageDump.end(); ++it){
    nestedGarbageDump.insert( (*it)->getNested() );
    delete *it;
  }

  // make an initial unsorted version of the array
  AssemblyJob** jobArray = new AssemblyJob*[ jobSet->size() + 1 ];
  long totalNumJobs = 0;
  for (set<AssemblyJob*>::iterator jobIt = jobSet->begin(); jobIt != jobSet->end(); ++jobIt){
    jobArray[totalNumJobs] = *jobIt;
    totalNumJobs++;
  }
  delete jobSet;

  // i want the job array to be sorted by size, so first i need to know
  // how many jobs of each size there are.  while i'm at it, i am also
  // adding normalization counts to the SSNormalized job components.
  // NOTE: because of buffering, this loop is not thread-able, despite how it may first appear
  map<long, long> sizeToJobCount;
  long* arrayOfJobSizes = new long[ totalNumJobs ];
  long totalOfJobSizes = 0;
  for (long oc = 0; oc < totalNumJobs; ++oc){
    set<ScoredSeq*> inputSeqs;
    jobArray[oc]->allInputSeqs( &inputSeqs );
    for (set<ScoredSeq*>::iterator seqIt = inputSeqs.begin(); seqIt != inputSeqs.end(); ++seqIt){
      dynamic_cast<ScoredSeqNormalized*>( *seqIt )->addCount();
      // this sequence will be deleted separately
      nestedGarbageDump.erase( *seqIt );
      // only delete it locally if it is not an input sequence
      if (_inputContigs.count(dynamic_cast<ScoredSeqNormalized*>(*seqIt))==0){
	_seqToTotalUseCount.insert( pair<ScoredSeq*,long>( *seqIt, 0 ) );
      }
    }

    // DEFINE the job size as a long
    // i will sort by the product of the biggest contig times the summed lengths of all the rest
    // jobs with just one contig are null, and they will be appropriately sized as zero by this metric
    // note: getting sizes will require buffering

    long jobSize = 0;
    if (inputSeqs.size() > 1){
      addBufferedSeqs( &inputSeqs );
      for (int n = 0; n < _numReadFiles; ++n){ ReadFileCommunicator::bufferFill(_readFileArray[n], ReadFile::bufferThreaded); }
      long totalContigSize = 0;
      long maxContigSize = 0;
      for (set<ScoredSeq*>::iterator seqIt = inputSeqs.begin(); seqIt != inputSeqs.end(); ++seqIt){
	long contigSize = (*seqIt)->size();
	totalContigSize += contigSize;
	if (contigSize > maxContigSize){ maxContigSize = contigSize; }
      }
      removeBufferedSeqs( &inputSeqs );
      if (maxContigSize != 0){
	long upperBound = LONG_MAX / maxContigSize;
	long remainingLength = totalContigSize - maxContigSize;
	if (remainingLength < upperBound){
	  jobSize = maxContigSize * (totalContigSize - maxContigSize);
	} else { jobSize = LONG_MAX; }
      }
    }

    arrayOfJobSizes[oc] = jobSize;
    totalOfJobSizes += jobSize;
    map<long,long>::iterator stjcIt = sizeToJobCount.find( jobSize );
    if (stjcIt == sizeToJobCount.end()){ sizeToJobCount.insert( pair<long,long>(jobSize,1) ); }
    else { stjcIt->second++; }
  }

  // delete the nested reads that are not part of jobs
  for (set<ScoredSeq*>::iterator it = nestedGarbageDump.begin(); it != nestedGarbageDump.end(); ++it){
    (*it)->deepDelete();
  }


  // SORT THE JOBS

  // create another data structure to make sorting of the list easier
  map<long, long> sizeToStartIndex;
  long currentStartIndex = totalNumJobs;
  for (map<long,long>::iterator it = sizeToJobCount.begin(); it != sizeToJobCount.end(); ++it){
    currentStartIndex -= it->second;
    sizeToStartIndex.insert( pair<long,long>(it->first, currentStartIndex) );
  }

  // now i can use the information that i gathered above to sort the array
  AssemblyJob** sortedJobArray = new AssemblyJob*[ totalNumJobs ];
  long* arrayOfSortedJobSizes = new long[ totalNumJobs ];
  for (long oc = 0; oc < totalNumJobs; ++oc){
    map<long,long>::iterator indexIt = sizeToStartIndex.find( arrayOfJobSizes[oc] );
    arrayOfSortedJobSizes[ indexIt->second ] = arrayOfJobSizes[oc];
    sortedJobArray[ indexIt->second ] = jobArray[oc];
    indexIt->second++;
  }

  // replace the unsorted with the sorted versions
  delete [] jobArray;
  jobArray = sortedJobArray;
  delete [] arrayOfJobSizes;
  arrayOfJobSizes = arrayOfSortedJobSizes;

  // organize the jobs into those to be threaded and those to not be threaded
  long remainingJobSizeTotal = totalOfJobSizes;
  long startParallelIndex = 0;
  int numThreads = omp_get_max_threads();
  // if there is only one thread, no jobs will be run in parallel
  if (numThreads == 1){ startParallelIndex = totalNumJobs; }
  else if (totalNumJobs > 0){
    remainingJobSizeTotal -= arrayOfJobSizes[startParallelIndex];
    while (startParallelIndex < totalNumJobs and arrayOfJobSizes[startParallelIndex] * numThreads >= remainingJobSizeTotal){
      startParallelIndex++;
      if (startParallelIndex < totalNumJobs){ remainingJobSizeTotal -= arrayOfJobSizes[startParallelIndex]; }
    }
  }
  if (totalNumJobs - startParallelIndex < numThreads){ startParallelIndex = totalNumJobs; }

  _listener->beginEdgeAssembly(totalNumJobs);

  // an output collection for each thread; this will be consolidated after
  set<ScoredSeq*>* metaInputByThread[ _numThreads ];
  for (int tn = 0; tn < _numThreads; ++tn){ metaInputByThread[tn] = new set<ScoredSeq*>; }


  // RUN THE JOBS

  // first run the long jobs individually threaded
  for (long oc = 0; oc < startParallelIndex; ++oc){
    runOneMiniAssembly(jobArray[oc], metaInputByThread[0], preMetaFilter, AssemblyJob::THREADED);
  }

  // now run the smaller jobs in parallel (biggest launched first)
  #pragma omp parallel for schedule(dynamic)
  for (long oc = startParallelIndex; oc < totalNumJobs; ++oc){
    int threadNum = omp_get_thread_num();
    runOneMiniAssembly(jobArray[oc], metaInputByThread[threadNum], preMetaFilter, AssemblyJob::NOTTHREADED);
  }

  // clean up array out here
  for (long oc = 0; oc < totalNumJobs; ++oc){ delete jobArray[oc]; }
  delete [] jobArray;
  delete [] arrayOfJobSizes;

  // consolidate the output and clean up
  for (int n = 0; n < _numThreads; ++n){
    _metaInputContigs.insert( metaInputByThread[n]->begin(), metaInputByThread[n]->end() );
    delete metaInputByThread[n];
  }

  _listener->endEdgeAssembly();
}





void ExtendCycle::runMetaAssembly(ParamsAlignment* pa, ParamsDeBruijn* pdb){
  _listener->beginMetaAssembly( _metaInputContigs.size() );
  AssemblyJobFactory * contigFactory = new AssemblyJobFactory(pa, pdb, AssemblyJob::DOUBLESTRANDED, _listener);
  AssemblyJob* aj = contigFactory->redundancyJob( &_metaInputContigs );

  //singleAssemblyHelper(aj, &_assembledContigs, AssemblyJob::threaded);

  aj->runJob(AssemblyJob::THREADED);
  aj->allAssembledSeqs( &_assembledContigs );
  // get the input data that was combined into the output contigs and delete it
  set<ScoredSeq*> discardedInput;
  aj->discardedInputSeqs( &discardedInput );
  for (set<ScoredSeq*>::iterator disIt = discardedInput.begin(); disIt != discardedInput.end(); ++disIt){
    (*disIt)->deepDelete();
  }

  delete aj;
  delete contigFactory;

  // delete all of the input contigs - this will include the _doNotExtendContigs
  // shallow delete since I am only deleting the shells
  for (set<ScoredSeqNormalized*>::iterator icIt = _inputContigs.begin(); icIt != _inputContigs.end(); ++icIt){ delete *icIt; }
  _inputContigs.clear();

  _listener->endMetaAssembly( _assembledContigs.size() );
}



bool ExtendCycle::cycleHasRun(){ return _cycleHasRun; }
void ExtendCycle::getContigs(set<ScoredSeq*>* outputContigs){
  outputContigs->insert(_assembledContigs.begin(), _assembledContigs.end());
}


#endif
